package cabler;

import java.util.ArrayList;
import java.util.Map;
import java.util.TreeMap;
import java.io.*;
import dragonfly.graph.Node;

/**
 * Generates list of edges for selected cabling method.
 * 
 * @3author Emily Hastings 
 * @version 7/16/13
 * 
 * Updated 8/12/13 to reflect new naming scheme (relative/absolute vs. local/global)
 * Updated 3/12/15 to add new method (circulant)
 * Updated 7/10/15 to add new method (1GCC) @ author Sofia Meyers
 * Updated 6/14/16 to add new method (Unity-alternating) @ author Myra Doubet and name of 1GCC changed to Unity-reset
 */

public class DragonflyCabler
{    
	/**
	 * routers num groups relative/absolute
	 * 
	 * Generates 4-tuples representing cable connections as follows:  (Group 1, Router 1, Group 2, Router 2)
	 * Takes as arguments number of routers per group, number of groups, and desired cabling method (relative, absolute, circulant, UnityR, or UnityA).
	 * Writes results to file called edgeList.txt, with routersPerGroup and numGroups on first line (separated by white space)
	 * followed by lines representing each tuple, with each component of the tuple separated by white space.     
	 */
	public static void main(String[] args)  {
		if (args.length < 3)  {
			System.out.println("Invalid arguments.");
		}

		try  {
			/* initialize variables */
			int routersPerGroup = Integer.parseInt(args[0]); // how many routers per group-0
			int numGroups = Integer.parseInt(args[1]);// how many groups there are
			int globalsPerRouter = (numGroups - 1)/routersPerGroup; // how many global edges per switch
			ArrayList<String> edgeList = new ArrayList<String>();
			String method = args[2];

			/* create file */
			FileWriter fstream = new FileWriter("edgeList.txt");
			BufferedWriter out = new BufferedWriter(fstream);
			out.write(routersPerGroup + " " + numGroups + "\n");

			//System.out.println(routersPerGroup + " " + numGroups);

			/* determine cabling */
			if (method.equals("relative"))  {
				
				relative(routersPerGroup, numGroups, globalsPerRouter, edgeList);
				
			} else if (method.equals("absolute")){
				
				absolute(routersPerGroup, numGroups, globalsPerRouter, edgeList);
				
			} else if(method.equals("circulant")) {
				
				circulant(routersPerGroup, numGroups, globalsPerRouter, edgeList);
				
			} else if(method.equals("nautilus")){
				
				nautilus(routersPerGroup, numGroups, globalsPerRouter, edgeList);
				
			} else if(method.equals("nautilusA")){
				
				nautilusA(routersPerGroup, numGroups, globalsPerRouter, edgeList);
				
		} else if(method.equals("helix")){
			
			helix(routersPerGroup, numGroups, globalsPerRouter, edgeList);
			
		} else if(method.equals("helix2")){
			
			helix2(routersPerGroup, numGroups, globalsPerRouter, edgeList);
			
		} else {
			System.out.println("Invalid arguments.");
			return;
		}

			/**write list to file*/
			for (String s : edgeList) {
				//System.out.println(s);
				out.write(s + "\n");
			}            
			out.close();

		} catch (Exception e)  {
			
			e.printStackTrace(System.err);
		}
	}
	
	public static void relative(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		for (int i=0; i<numGroups; i++)  {
			for (int j=0; j<routersPerGroup; j++)  {
				for (int k=1; k<=globalsPerRouter; k++) {
					int g1 = i;
					int r1 = j;
					int g2 = (i + (globalsPerRouter * j) + k) % numGroups;
					int r2 = (routersPerGroup - 1) - j;

					String retVal = g1 + " " + r1 + " " + g2 + " " + r2;

					//check for duplicate edges
					if (!edgeList.contains(g2 + " " + r2 + " " + g1 + " " + r1))  {
						edgeList.add(retVal);
					}
				}
			}    
		}
	}
	
	public static void absolute(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		for (int i=0; i<numGroups; i++)  {
			boolean skipped = false;
			for (int j=0; j<routersPerGroup; j++)  {
				for (int k=0; k<globalsPerRouter; k++)  {
					int g1 = i;
					int r1 = j;
					int g2 = (globalsPerRouter * j) + k;
					if ((g2 == g1) || skipped == true)  {  //compensate for omitting own group
						g2 = g2 + 1;
						skipped = true;
					}
					if (g2 < g1)  {  //all connections to previous groups have already been counted
						continue;
					}
					int r2 = i / globalsPerRouter;

					String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               

					//already removed duplicates, so just add to list
					edgeList.add(retVal);
				}
			}
		}
	}
	
	public static void circulant(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		if(globalsPerRouter % 2 != 0) {
			System.err.println("Can only use circulant scheme with an even number of global edges per router");
			System.exit(1);
		}
		for (int i=0; i<numGroups; i++)  {
			for (int j=0; j<routersPerGroup; j++)  {
				for (int k=0; k<globalsPerRouter/2; k++)  {
					int g1 = i;
					int r1 = j;
					int diff = j*globalsPerRouter/2 + (k+1);
					int g2 = (i+diff) % numGroups;
					int r2 = j;
					if (g2 > g1)  {  //only add if it connects to higher group
						String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               
						edgeList.add(retVal);
					}
					g2 = (i-diff+numGroups) % numGroups;
					if (g2 > g1)  {  //only add if it connects to higher group
						String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               
						edgeList.add(retVal);
					}

				}
			}
		}
	}
	
	public static void nautilus(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		boolean [][] connectedGr=new boolean [numGroups][numGroups];// array to know which groups you are connected in order to skip them
		
		// initialize the boolean array to false at the beginning since we are not connected to any groups yet
		// except when there is a connection to yourself which needs to be true since you can't connect to yourself
		int [] switchCount = new int [(numGroups*routersPerGroup)];// array to keep count of how many connections each switch yes
		for(int i=0;i<numGroups;i++){
			for(int j=0;j<numGroups;j++){
				if(i==j){
					connectedGr[i][j]=true;
				}else{
					connectedGr[i][j]=false;
				}
			}
		}
		// go through each group
		for(int k=0;k<numGroups;k++){
			// go through each switch 
			for(int m=0;m<routersPerGroup;m++){ 
				int g1=k;
				int r1=m;
				// calculate which switch to connect to will be the same for the whole group
				int r2= g1%routersPerGroup; // need to get the current group 
				// figure out how many connection you will have to make
				int howManyConnections=globalsPerRouter-switchCount[(g1*routersPerGroup)+r1];
				int h=m;
				while(howManyConnections!=0){
					do {
						if(r1%2==0)
							h = (h+1) % numGroups;  // even switches means you go positive direction
						else
							h= (h-1+numGroups) % numGroups; // odd switches means you go in the negative direction
					} while(connectedGr[g1][h]); // find the next free group
					//connect to group h
					String NextLink = g1 + " " + r1 + " " + h + " " + r2;
					// update the boolean array to reflect new Connection (edges are 2 way directed edges) 
					connectedGr[g1][h]= true;
					connectedGr[h][g1]=true;
					//update switch count to reflect new Connection (edges are 2 way directed edges) 
					switchCount[(g1*routersPerGroup)+(r1)]=switchCount[(g1*routersPerGroup)+(r1)]+1;
					switchCount[(h*routersPerGroup)+(r2)]=switchCount[(h*routersPerGroup)+(r2)]+1;
					//add the link to the text file
					edgeList.add(NextLink);
					//System.out.println(NextLink);
					howManyConnections--;

				}
			}
		}
	}
	
	public static void nautilusA(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		boolean [][] connectedGr=new boolean [numGroups][numGroups];// array to know which groups you are connected in order to skip them
		
		// initialize the boolean array to false at the beginning since we are not connected to any groups yet
		// except when there is a connection to yourself which needs to be true since you can't connect to yourself
		int [] switchCount = new int [(numGroups*routersPerGroup)];// array to keep count of how many connections each switch yes
		for(int i=0;i<numGroups;i++){
			for(int j=0;j<numGroups;j++){
				if(i==j){
					connectedGr[i][j]=true;
				}else{
					connectedGr[i][j]=false;
				}
			}
		}
		// go through each group
		for(int k=0;k<numGroups;k++){
			// go through each switch 
			for(int m=0;m<routersPerGroup;m++){ 
				int g1=k;
				int r1=m;
				// calculate which switch to connect to will be the same for the whole group
				int r2= g1%routersPerGroup; // need to get the current group 
				// figure out how many connection you will have to make
				int howManyConnections=globalsPerRouter-switchCount[(g1*routersPerGroup)+r1];
				int h=m;
				while(howManyConnections!=0){
					do {
						if(routersPerGroup%2==0){  //if you have an even number of groups, always start with positive direction
							if(r1%2==0)
								h = (h+1) % numGroups;  // even switches means you go positive direction
							else
								h= (h-1+numGroups) % numGroups; // odd switches means you go in the negative direction
						} else {  //if you have an odd number of groups
							if(g1%2==0){  //if the group number is even, start with positive direction
								if(r1%2==0)
									h = (h+1) % numGroups;  // even switches means you go positive direction
								else
									h= (h-1+numGroups) % numGroups; // odd switches means you go in the negative direction
							} else {  // if the group number is odd, start with negative direction
								if((r1+1)%2==0)
									h = (h+1) % numGroups;  // odd switches means you go positive direction
								else
									h= (h-1+numGroups) % numGroups; // even switches means you go in the negative direction
							}
						}
					} while(connectedGr[g1][h]); // find the next free group
					//connect to group h
					String NextLink = g1 + " " + r1 + " " + h + " " + r2;
					// update the boolean array to reflect new Connection (edges are 2 way directed edges) 
					connectedGr[g1][h]= true;
					connectedGr[h][g1]=true;
					//update switch count to reflect new Connection (edges are 2 way directed edges) 
					switchCount[(g1*routersPerGroup)+(r1)]=switchCount[(g1*routersPerGroup)+(r1)]+1;
					switchCount[(h*routersPerGroup)+(r2)]=switchCount[(h*routersPerGroup)+(r2)]+1;
					//add the link to the text file
					edgeList.add(NextLink);
					//System.out.println(NextLink);
					howManyConnections--;

				}
			}
		}
	}
	
	public static void helix(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		for (int i=0; i<numGroups; i++)  {
			for (int j=0; j<routersPerGroup; j++)  {
				for(int k=0; k<(globalsPerRouter/2); k++){
					int g1 = i;
					int r1 = j;
					int g2 = (i + j*(globalsPerRouter/2) + k + 1)%numGroups;
					int r2 = (j+1)%routersPerGroup;
					String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               

					//doesn't duplicate
					edgeList.add(retVal);
				}
				//Adding the extra edges if odd
				if(globalsPerRouter%2==1){
					int g1 = i;
					int r1 = j;
					int g2 = (i+ (globalsPerRouter/2)*routersPerGroup + j + 1)%numGroups;
					int r2 = (routersPerGroup-j-1)%routersPerGroup;
					if(g1<g2){
						String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               

					//doesn't duplicate
						edgeList.add(retVal);
					}
				}
			}
		}
	}
	
	public static void helix2(int routersPerGroup, int numGroups, int globalsPerRouter, ArrayList<String> edgeList){
		for (int i=0; i<numGroups; i++)  {
			for (int j=0; j<routersPerGroup; j++)  {
				for(int k=0; k<(globalsPerRouter/2); k++){
					int g1 = i;
					int r1 = j;
					int g2 = (i + k*routersPerGroup + j + 1)%numGroups;
					int r2 = (j+1)%routersPerGroup;
					String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               

					//doesn't duplicate
					edgeList.add(retVal);
				}
				//Adding the extra edges if odd
				if(globalsPerRouter%2==1){
					int g1 = i;
					int r1 = j;
					int g2 = (i+ (globalsPerRouter/2)*routersPerGroup + j + 1)%numGroups;
					int r2 = (routersPerGroup-j-1)%routersPerGroup;
					if(g1<g2){
						String retVal = g1 + " " + r1 + " " + g2 + " " + r2;               

					//doesn't duplicate
						edgeList.add(retVal);
					}
				}
			}  
		}
	}
}
