package dragonfly.graph;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Scanner;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;
import dragonfly.DFNameComparator;

/**
 * Adapted from Jaime Spacco's graph library
 * 
 */
public class Graph {
	private Map<String, Node> nodes = new TreeMap<String, Node>();

	/**
	 * number of nodes in the graph
	 * @return
	 */
	public int size(){
		return nodes.size();
	}
	
	public void addNode(Node n) {
		nodes.put(n.getName(), n);
	}

	public void breadthFirstSearch(String startNodeName, NodeVisitor v) {
		// queue of nodes still to visit
		Queue<Node> toVisit = new LinkedList<Node>();
		// list of nodes already visited
		Set<Node> visited = new HashSet<Node>();
		
		// start node
		Node start = getOrCreateNode(startNodeName);
		System.out.println("start: " + start);
		System.out.println();
		toVisit.add(start);

		while (toVisit.size() > 0) {
			// take from front of queue
			Node current = toVisit.remove();
	
			if (!visited.contains(current)) {
				v.visit(current);
			}
			visited.add(current);

			for (Node n : current.getNeighbors()) {
				if (!visited.contains(n)) {
					// append onto the end of the queue
					// (toVisit is a queue)
					toVisit.add(n);
				}
			}
		}

	}

	public void depthFirstSearch(String startNodeName, NodeVisitor v) {
		// stack of nodes still to visit
		Stack<Node> toVisit = new Stack<Node>();
		// list of nodes already visited
		Set<Node> visited = new HashSet<Node>();

		// start node
		Node start = nodes.get(startNodeName);
		toVisit.add(start);

		while (toVisit.size() > 0) {
			// pop from beginning of the stack
			Node current = toVisit.pop();
			if (!visited.contains(current)) {
				// System.out.println(current);
				v.visit(current);
			}
			visited.add(current);

			for (Node n : current.getNeighbors()) {
				if (!visited.contains(n)) {
					// push onto the stack
					// (toVisit is a queue)
					toVisit.push(n);
				}
			}
		}
	}

	private static Collection<Path> getEdges(Node node, float offset) {
		List<Path> list = new LinkedList<Path>();
		for (Node n : node.getNeighbors()) {
			Path p = new Path(n, node.getWeight(n) + offset);
			list.add(p);
		}
		return list;
	}

	public Map<Node, Float> dijkstra(String startName) {
		Map<Node, Float> map = new HashMap<Node, Float>();
		Node n = getOrCreateNode(startName);
		map.put(n, 0f);
		PriorityQueue<Path> queue = new PriorityQueue<Path>();
		queue.addAll(getEdges(n, 0));
		while (!queue.isEmpty()) {
			Path p = queue.poll();
			if (!map.containsKey(p.getDst())) {
				map.put(p.getDst(), p.getCost());
			}
			for (Path x : getEdges(p.getDst(), map.get(p.getDst()))) {
				if (!map.containsKey(x.getDst())) {
					queue.add(x);
				}
			}
		}
		return map;
	}
	
	public void path(String goal) {
		Map<Node, Float> m = dijkstra(goal);
		for(Map.Entry<Node, Float> entry : m.entrySet()) {
			Node curr = entry.getKey();
			float dist = entry.getValue();
			while(!curr.getName().equals(goal)) {
				System.out.print(curr + "  ");
				for(Node n: curr.getNeighbors()) {
					if(m.get(n) < dist - 0.5) {
						curr = n;
						dist = dist - 1;
					}
				}
			}
			System.out.println(goal + "  " + entry.getValue());
//		    System.out.println(entry.getKey() + "/" + entry.getValue());
//		    System.out.println(entry.getKey().getNeighbors());
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see graphs.IGraph#primJarnik()
	 */
	public Set<Edge> primJarnik() {
		Set<Edge> results = new LinkedHashSet<Edge>();
		Set<String> finished = new HashSet<String>();

		TreeSet<Edge> frontier = new TreeSet<Edge>();

		// Grab the first node we can find
		Collection<Node> nodes = getAllNodes();
		Node current = (Node) nodes.toArray()[0];
		finished.add(current.getName());
		while (finished.size() < nodes.size()) {
			for (Node neighbor : current.getNeighbors()) {
				Edge edge = new Edge(current, neighbor,
						current.getWeight(neighbor));
				// frontier.add(edge);
				if (!finished.contains(edge.getSrc().getName())
						|| !finished.contains(edge.getDst().getName())) {
					frontier.add(edge);
				}
			}
			Edge e = null;
			while (true) {
				e = frontier.first();
				frontier.remove(e);
				if (!finished.contains(e.getSrc().getName())
						|| !finished.contains(e.getDst().getName())) {
					break;
				}
			}

			results.add(e);
			// System.out.println("adding "+e);
			// since sets only store one copy, just add both src and dst
			finished.add(e.getSrc().getName());
			finished.add(e.getDst().getName());
			Node n1 = e.getSrc();
			if (finished.contains(n1.getName())) {
				n1 = e.getDst();
			}
			current = n1;
		}
		return results;
	}

	public Set<Edge> modifiedPrimJarnik(String firstNodeName) {
		PriorityQueue<Edge> edgeSet = new PriorityQueue<Edge>();
		HashSet<Node> visitedNodeSet = new HashSet<Node>();
		Set<Edge> returnEdgeSet = new LinkedHashSet<Edge>();
		
		Node firstNode = null;
		firstNode = nodes.get(firstNodeName);
		visitedNodeSet.add(firstNode);

		for (Node n : firstNode.getNeighbors()) {
			edgeSet.add(new Edge(firstNode, n, firstNode.getWeight(n)));
		}

		while (!edgeSet.isEmpty()) {
			Edge topEdge = edgeSet.poll();
			Node candidateVisit = topEdge.getDst();

			if (visitedNodeSet.contains(candidateVisit))
				continue;
			else {
				returnEdgeSet.add(topEdge);
				visitedNodeSet.add(candidateVisit);
				for (Node n : candidateVisit.getNeighbors()) {
					edgeSet.add(new Edge(candidateVisit, n, candidateVisit
							.getWeight(n)));
				}
			}

		}
		return returnEdgeSet;
	}

	/**
	 * Finds number of connected components and the sizes of each component
	 * This is the idea of 'global gangs'
	 * @return list of component sizes
	 */
	public ArrayList<Integer> connectedComponentSizes() {
		Graph temp = this.clone();
		ArrayList<Integer> sizes = new ArrayList<Integer>();
		Collection<Node> unvisitedNodes = temp.getAllNodes();
		while (!unvisitedNodes.isEmpty()) {
			Node firstNode = unvisitedNodes.iterator().next();
			Set<Edge> tempset = temp.modifiedPrimJarnik(firstNode.getName());
			for (Edge e : tempset) {
				unvisitedNodes.remove(e.getSrc());
				unvisitedNodes.remove(e.getDst());
			}
			sizes.add(tempset.size() + 1);
		}
		return sizes;
	}
	
	/**
	 * Finds all the connected components and the members of each
	 * @return a list of lists of all the members in each connectedComponent
	 */
	public ArrayList<ArrayList<Node>> connectedComponentsList(){
		Graph tempGraph = this.clone();
		ArrayList<ArrayList<Node>> connectedCompList = new ArrayList<ArrayList<Node>>();
		Collection<Node> unvisitedNodes = tempGraph.getAllNodes();
		while (!unvisitedNodes.isEmpty()) {
			ArrayList<Node> tempSet = new ArrayList<Node>();
			Node firstNode = unvisitedNodes.iterator().next();
			Set<Edge> temp = tempGraph.modifiedPrimJarnik(firstNode.getName());
			for (Edge e: temp){
				unvisitedNodes.remove(e.getSrc());
				if(!tempSet.contains(e.getSrc())){
					tempSet.add(e.getSrc());
				}
				unvisitedNodes.remove(e.getDst());
				if(!tempSet.contains(e.getDst())){
					tempSet.add(e.getDst());
				}
			}
			connectedCompList.add(tempSet);
		}
		return connectedCompList;
	}
	
	/**
	 * Finds all the connected components and the members of each
	 * @return a list of lists of all the members in each connectedComponent
	 */
	public ArrayList<ArrayList<Edge>> connectedComponentsEdgesList(){
		ArrayList<ArrayList<Node>> nodesList = connectedComponentsList();
		ArrayList<ArrayList<Edge>> edgeList = new ArrayList<ArrayList<Edge>>();
		for(ArrayList<Node> nodes : nodesList){
			ArrayList<Edge> tempSet = new ArrayList<Edge>();
			for(Node n : nodes){
				for(Edge e : n.getAllEdges()){
					if(!tempSet.contains(e)){ 
						// does compare weights as well as names
						//weights should be the same
						tempSet.add(e);
					}
				}
				
			}
			edgeList.add(tempSet);
		}
		return edgeList;
	}
	
	
	
	
	public Node getOrCreateNode(String name) {
		if (!nodes.containsKey(name)) {
			nodes.put(name, new Node(name));
		}
		return nodes.get(name);
	}

	public Node getNode(String name) {
		return nodes.get(name);
	}

	public boolean containsNode(String name) {
		return nodes.containsKey(name);
	}

	public Collection<Node> getAllNodes() {
		return nodes.values();
	}

	public String toCsv() {
		StringBuilder str = new StringBuilder();
		str.append("node names\n");
		for (Node n : getAllNodes()) {
			str.append(n.getName());
			str.append("\n");
		}
		str.append("connections\n");
		for (Node n : getAllNodes()) {
			for (Node nn : n.getNeighbors()) {
				str.append(n.getName());
				str.append(",");
				str.append(nn.getName());
				str.append(",");
				str.append(n.getWeight(nn));
				str.append("\n");
			}
		}
		return str.toString();
	}

	/**
	 * Give it a CSV file created by saving the result of the toCsv method
	 * 
	 * @param filename
	 * @return Null if for any reason it failed
	 */
	public static Graph fromCsv(String filename) {
		Scanner sc;
		try {
			sc = new Scanner(new File(filename));
		} catch (Exception e) {
			return null;
		}
		return fromCsv(sc);
	}

	private static Graph fromCsv(Scanner sc) {
		Graph g = new Graph();
		try {
			sc.nextLine(); // node name line
			String str = sc.nextLine();
			while (!str.equals("connections")) {
				g.addNode(new Node(str));
				str = sc.nextLine();
			}
			String[] connection;
			while (sc.hasNextLine()) {
				connection = sc.nextLine().split(",");
				g.getNode(connection[0]).addDirectedEdgeToNode(
						g.getNode(connection[1]),
						Float.parseFloat(connection[2]));
			}
		} catch (Exception e) {
			return null;
		}
		return g;
	}	

	/**
	 * Deep copy
	 */
	public Graph clone() {
		String csv = this.toCsv(); // serialize the graph
		Scanner sc = new Scanner(csv);
		Graph copy = fromCsv(sc); // then unserializes it
		sc.close();
		return copy;
	}
	
	/**
	 * Given a list of nodes, finds all the edges
	 * which have both endpoints within the list
	 * @param nodeList
	 * @return list of edges
	 */
	public ArrayList<Edge> edgesBetweenNode(ArrayList<Node> nodeList){
		ArrayList<Edge> edgeList = new ArrayList<Edge>();
		DFNameComparator comp = new DFNameComparator();
		for(Node v : nodeList){
			for(Node u : v.getNeighbors()){
				if(nodeList.contains(u) // compare them to make sure we don't double count
						&& comp.compare(u.getName(), v.getName()) < 0)
					edgeList.add(new Edge(v,u,v.getWeight(u)));
			}
		}	
		return edgeList;
	}

}